import sys
import functools
import pandas as pd
from PAMI.fuzzyCorrelatedPattern.basic.abstract import *

class FFList:
    """
     A class represent a Fuzzy List of an element

    Attributes :
    ----------
         item: int
             the item name
         sumIutil: float
             the sum of utilities of an fuzzy item in database
         sumRutil: float
             the sum of resting values of a fuzzy item in database
         elements: list
             a list of elements contain tid,Utility and resting values of element in each transaction
    Methods :
    -------
        addElement(element)
            Method to add an element to this fuzzy list and update the sums at the same time.

        printelement(e)
            Method to print elements            

    """

    def __init__(self, itemName, region):
        self.item = itemName
        self.region = region
        self.sumIutil = 0.0
        self.sumRutil = 0.0
        self.elements = []

    def addElement(self, element):
        """
            A Method that add a new element to FFList

            :param element: an element to be add to FFList
            :pram type: Element
        """
        self.sumIutil += element.iutils
        self.sumRutil += element.rutils
        self.elements.append(element)


class Element:
    """
        A class represents an Element of a fuzzy list

    Attributes :
    ----------
        tid : int
            keep tact of transaction id
        iutils: float
            the utility of an fuzzy item in the transaction
        rutil : float
            the neighbourhood resting value of an fuzzy item in the transaction
    """

    def __init__(self, tid, iutil, rutil):
        self.tid = tid
        self.iutils = iutil
        self.rutils = rutil


class Reagions:
    """
            A class calculate the region value

    Attributes:
    ----------
            low : int
                low region value
            middle: int 
                middle region value
            high : int
                high region values
        """

    def __init__(self, item, quantity, regionsNumber, mapOfregios):
        self.low = 0
        self.middle = 0
        self.high = 0
        if regionsNumber == 3:
            if 0 < quantity <= 1:
                self.low = 1
                self.high = 0
                self.middle = 0
                t1 = (item, 'L')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1
            elif 1 <= quantity < 6:
                self.low = float((-0.2 * quantity) + 1.2)
                self.middle = float((0.2 * quantity) - 0.2)
                self.high = 0
                t1 = (item, 'L')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1
                t1 = (item, 'M')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1
            elif 6 <= quantity <= 11:
                self.low = 0
                self.middle = float((-0.2 * quantity) + 2.2)
                self.high = float((0.2 * quantity) - 1.2)
                t1 = (item, 'M')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1
                t1 = (item, 'H')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1

            else:
                self.low = 0
                self.middle = 0
                self.high = 1
                t1 = (item, 'H')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios = temp + 1


class Pair:
    def __init__(self):
        """
            A Class to Store item and its quantity together
        """
        self.item = 0
        self.quantity = 0
        self.region = 'N'


class fcfp(corelatedFuzzyFrequentPatterns):
    """
        fcfp is the algorithm to discover Corelated Fuzzy-frequent patterns in a transactional database.
        it is based on traditianl fuzzy frequent pattern mining.

    Attributes :
    ----------
        iFile : file
            Name of the input file to mine complete set of fuzzy spatial frequent patterns
        oFile : file
               Name of the oFile file to store complete set of fuzzy spatial frequent patterns
        minSup : int
                The user given support
        minAllConf: float
             user Specified minAllConf( should be in range 0 and 1)
        memoryRSS : float
                To store the total amount of RSS memory consumed by the program
        startTime:float
               To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        itemsCnt: int
            To record the number of fuzzy spatial itemsets generated
        mapItemsLowSum: map
            To keep track of low region values of items
        mapItemsMidSum: map
            To keep track of middle region values of items
        mapItemsHighSum: map
            To keep track of high region values of items
        mapItemSum: map
            To keep track of sum of Fuzzy Vlues of items
        mapItemRegions: map
            To Kepp track of fuzzy regions of item
        jointCnt: int
            To keep track of the number of FFI-list that was constructed
        BufferSize: int
            represent the size of Buffer
        itemBuffer list
            to kepp track of items in buffer
    Methods :
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function            
        getRatio(self, prefix, prefixLen, item)
            Method to calculate the ration of itemset
        convert(value):
            To convert the given user specified value  
        FSFIMining( prefix, prefixLen, fsFim, minSup)
            Method generate FFI from prefix
        construct(px, py)
            A function to construct Fuzzy itemset from 2 fuzzy itemsets
        findElementWithTID(ulist, tid)
            To find element with same tid as given
        WriteOut(prefix, prefixLen, item, sumIutil,ratio)
            To Store the patten      

    Executing the code on terminal :
    -------
            Format: python3 fcfp.py <inputFile> <outputFile> <minSup> <minAllConf> <sep>
            Examples: python3 fcfp.py sampleTDB.txt output.txt 2 0.2 (minSup will be considered in support count or frequency)
                      python3 fcfp.py sampleTDB.txt output.txt 0.25 0.2 (minSup and maxPer will be considered in percentage of database)
                                                                     (it will consider separator as "\t")
                      python3 fcfp.py sampleTDB.txt output.txt 2 0.2 ,                    
                                                                      (it will consider separator as ',')
    Sample run of importing the code:
    -------------------------------
        
        import fcfp as alg

        obj = alg.fcfp("input.txt",2,0.4)

        obj.startMine()

        corelatedFuzzyFrequentPatterns = obj.getPatterns()

        print("Total number of Corelated Fuzzy Frequent Patterns:", len(corelatedFuzzyFrequentPatterns))

        obj.storePatternsInFile("outp")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
            The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.

    """
    startTime = float()
    endTime = float()
    minSup = str()
    maxPer = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    minAllConf=0.0
    sep="\t"
    memoryUSS = float()
    memoryRSS = float()

    def __init__(self, iFile, minSup, ratio,sep="\t"):
        super().__init__(iFile, minSup,ratio,sep)
        self.temp = {}
        self.mapItemRegionSum = {}
        self.start = 0
        self.end = 0
        self.itemsCnt = 0
        self.mapItemsLowSum = {}
        self.mapItemsMidSum = {}
        self.mapItemsHighSum = {}
        self.mapItemSum = {}
        self.mapItemRegions = {}
        self.joinsCnt = 0
        self.BufferSize = 200
        self.itemsetBuffer = []
        self.finalPatterns = {}
        self.dbLen = 0

    def compareItems(self, o1, o2):
        """
            A Function that sort all FFI-list in asendng order of Support
        """
        compare = self.mapItemSum[o1.item] - self.mapItemSum[o2.item]
        if compare == 0:
            return o1.item - o2.item
        else:
            return compare

    def findElementWithTID(self, ulist, tid):
        """
        To find element with same tid as given
        :param ulist: fuzzylist
        :type ulist: FFI-List
        :param tid: transaction id
        :type tid: int
        :return: element eith tid as given
        :rtype: element if exists or None
        """
        List = ulist.elements
        first = 0
        last = len(List) - 1
        while first <= last:
            mid = (first + last) >> 1
            if List[mid].tid < tid:
                first = mid + 1
            elif List[mid].tid > tid:
                last = mid - 1
            else:
                return List[mid]
        return None

    def convert(self, value):
        """
        To convert the given user specified value
        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self.dbLen * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self.dbLen * value)
            else:
                value = int(value)
        return value

    def startMine(self):
        """ 
            Frequent pattern mining process will start from here
        """
        self.start = time.time()
        self.minSup = self.convert(self.minSup)
        minSup = self.minSup
        with open(self.iFile, 'r') as file:
            for line in file:
                self.dbLen += 1
                parts = line.split(":")
                items = parts[0].split(self.sep)
                quanaities = parts[2].split(self.sep)
                for i in range(0, len(items)):
                    item = items[i]
                    regions = Reagions(item, int(quanaities[i]), 3, self.mapItemRegionSum)
                    if item in self.mapItemsLowSum.keys():
                        low = self.mapItemsLowSum[item]
                        low += regions.low
                        self.mapItemsLowSum[item] = low
                    else:
                        self.mapItemsLowSum[item] = regions.low
                    if item in self.mapItemsMidSum.keys():
                        mid = self.mapItemsMidSum[item]
                        mid += regions.middle
                        self.mapItemsMidSum[item] = mid
                    else:
                        self.mapItemsMidSum[item] = regions.middle
                    if item in self.mapItemsHighSum.keys():
                        high = self.mapItemsHighSum[item]
                        high += regions.high
                        self.mapItemsHighSum[item] = high
                    else:
                        self.mapItemsHighSum[item] = regions.high
            listOfFFIlist = []
            mapItemsToFFLIST = {}
            for item1 in self.mapItemsLowSum.keys():
                item = item1
                region = 'N'
                low = self.mapItemsLowSum[item]
                mid = self.mapItemsMidSum[item]
                high = self.mapItemsHighSum[item]
                if low >= mid and low >= high:
                    self.mapItemSum[item] = low
                    self.mapItemRegions[item] = "L"
                    region = 'L'
                elif mid >= low and mid >= high:
                    self.mapItemSum[item] = mid
                    self.mapItemRegions[item] = "M"
                    region = 'M'
                elif high >= low and high >= mid:
                    self.mapItemRegions[item] = "H"
                    region = 'H'
                    self.mapItemSum[item] = high
                if self.mapItemSum[item] >= self.minSup:
                    fuList = FFList(item, region)
                    mapItemsToFFLIST[item] = fuList
                    listOfFFIlist.append(fuList)
            listOfFFIlist.sort(key=functools.cmp_to_key(self.compareItems))
        tid = 0
        with open(self.iFile, 'r') as file:
            for line in file:
                parts = line.split(":")
                items = parts[0].split(self.sep)
                quanaities = parts[2].split(self.sep)
                revisedTransaction = []
                for i in range(0, len(items)):
                    pair = Pair()
                    pair.item = items[i]
                    regions = Reagions(pair.item, int(quanaities[i]), 3, self.temp)
                    item = pair.item
                    if self.mapItemSum[item] >= minSup:
                        if self.mapItemRegions[pair.item] == "L":
                            pair.quantity = regions.low
                            pair.region = 'L'
                        elif self.mapItemRegions[pair.item] == "M":
                            pair.region = 'M'
                            pair.quantity = regions.middle
                        elif self.mapItemRegions[pair.item] == "H":
                            pair.quantity = regions.high
                            pair.region = 'H'
                        if pair.quantity > 0:
                            revisedTransaction.append(pair)
                revisedTransaction.sort(key=functools.cmp_to_key(self.compareItems))
                for i in range(len(revisedTransaction) - 1, -1, -1):
                    pair = revisedTransaction[i]
                    remainUtil = 0
                    for j in range(len(revisedTransaction) - 1, i - 1, -1):
                        remainUtil += revisedTransaction[j].quantity
                    if pair.quantity > remainUtil:
                        remaingUtility = pair.quantity
                    else:
                        remaingUtility = remainUtil
                    if mapItemsToFFLIST.get(pair.item) is not None:
                        FFListOfItem = mapItemsToFFLIST[pair.item]
                        element = Element(tid, pair.quantity, remaingUtility)
                        FFListOfItem.addElement(element)
                tid += 1
        self.FSFIMining(self.itemsetBuffer, 0, listOfFFIlist, self.minSup)
        self.endtime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("No.Of Itemssets ", self.itemsCnt)
        print("joinsCnt: ", self.joinsCnt)

    def FSFIMining(self, prefix, prefixLen, FSFIM, minsup):
        """
            Generates FFSI from prefix

        :param prefix: the prefix patterns of FFSI
        :type prefix: len
        :param prefixLen: the length of prefix
        :type prefixLen: int
           :param FSFIM: the Fuzzy list of prefix itemsets
           :type FSFIM: list
           :param minsup: the minimum support of 
           :type minsup:int
           :param itemNighbours : the set of common neighbours of prefix
           :type itemNighbours : set
        """
        for i in range(0, len(FSFIM)):
            X = FSFIM[i]
            if X.sumIutil >= minsup:
                ratio = self.getRatio(prefix, prefixLen, X)
                if ratio >= self.minAllConf:
                    self.WriteOut(prefix, prefixLen, X, ratio)
            if X.sumRutil >= minsup:
                exULs = []
                for j in range(i + 1, len(FSFIM)):
                    Y = FSFIM[j]
                    exULs.append(self.construct(X, Y))
                    self.joinsCnt += 1
                self.itemsetBuffer.insert(prefixLen, X)
                self.FSFIMining(self.itemsetBuffer, prefixLen + 1, exULs, minsup)

    def construct(self, px, py):
        """
            A function to construct a new Fuzzy itemset from 2 fuzzy itemsets

            :param px:the itemset px
            :type px:FFI-List
            :param py:ithemset py
            :type py:FFI-List
            :return :the itemset of pxy(px and py)
            :rtype :FFI-List
        """
        pxyUL = FFList(py.item, py.region)
        for ex in px.elements:
            ey = self.findElementWithTID(py, ex.tid)
            if ey is None:
                continue
            eXY = Element(ex.tid, min([ex.iutils, ey.iutils], key=lambda x: float(x)), ey.rutils)
            pxyUL.addElement(eXY)
        return pxyUL

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
       """
        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
       """
        return self.endtime - self.start

    def getRatio(self, prefix, prefixLen, item):
        """Method to calculate the ration of itemset
            :param prefix: prefix of itemset
            :type prefix: list
            :param prefixLen: length of prefix
            :type prefixLen: int
            :param item: the last item
            :type item: int
            :return : corelated ratio
            :rtype: float
        """
        res = 1.0
        n = prefixLen
        for i in prefix:
            if self.mapItemRegionSum.get((i.item, i.region)) is not None and res < self.mapItemRegionSum[
                (i.item, i.region)]:
                res = self.mapItemRegionSum[(i.item, i.region)]
        if self.mapItemRegionSum.get((item.item, item.region)) is not None and res < self.mapItemRegionSum[
            (item.item, item.region)]:
            res = self.mapItemRegionSum[(item.item, item.region)]
        return item.sumIutil / res

    def WriteOut(self, prefix, prefixLen, item, ratio):
        """
            To Store the patten
            :param prefix: prefix of itemset
            :type prefix: list
            :param prefixLen: length of prefix
            :type prefixLen: int
            :param item: the last item
            :type item: int
            :param sumIutil: sum of utility of itemset
            :type sumIutil: float
            :param ratio: the ratio of itemset
            :type ratio: float
        """
        self.itemsCnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i].item) + "." + str(prefix[i].region) + ' '
        res += str(item.item) + "." + str(item.region)
        res1 = str(item.sumIutil) + " : " + str(ratio) + "\n"
        # self.bwriter.write(res)
        self.finalPatterns[res] = res1

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataframe = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = str(x) + " : " + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6: # to includes separator
            ap = fcfp(sys.argv[1], sys.argv[3],float(sys.argv[4]),sys.argv[5])
        if len(sys.argv) == 5: # to consider '\t' as separator
           ap = fcfp(sys.argv[1], sys.argv[3],float(sys.argv[4]))
        ap.startMine()
        fuzzycorelatedFrequentPattenrs = ap.getPatterns()
        print("Total number of Fuzzy-Frequent Patterns:", len(fuzzycorelatedFrequentPattenrs))
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in seconds:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")


from abc import ABC, abstractmethod
import time
import csv
import pandas as pd
from collections import defaultdict
from itertools import combinations as c
import os
import os.path
import psutil
from array import *
import datetime
import resource
import functools
class utilityPatterns(ABC):
    """ This abstract base class defines the variables and methods that every utility pattern mining algorithm must
    employ in PAMI

        ...

        Attributes
        ----------
        iFile : str
            Input file name or path of the input file
        minUtil: float
            UserSpecified minimum utility value.
        startTime:float
            To record the start time of the algorithm
        endTime:float
            To record the completion time of the algorithm
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        oFile : str
            Name of the output file to store complete set of utility patterns
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

        Methods
        -------
        startMine()
            Mining process will start from here
        getUtilityPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of utility patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of utility patterns will be loaded in to data frame
        getMemoryUSS()
            Total amount of USS memory consumed by the program will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the program will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the program will be retrieved from this function
    """

    def __init__(self, iFile,nFile,minUtil):
        """
        :param iFile: Input file name or path of the input file
        :type iFile: str
        :param minUtil: UserSpecified minimum utility value.
        :type minUtil: float
        """

        self.iFile = iFile
        self.nFile=nFile
        self.minUtil = minUtil

    @abstractmethod
    def iFile(self):
        """Variable to store the input file path/file name"""

        pass

    @abstractmethod
    def minUtil(self):
        """Variable to store the user-specified minimum utility value"""

        pass

    @abstractmethod
    def startTime(self):
        """Variable to store the start time of the mining process"""

        pass

    @abstractmethod
    def endTime(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryUSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryRSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def finalPatterns(self):
        """Variable to store the complete set of patterns in a dictionary"""

        pass

    @abstractmethod
    def oFile(self):
        """Variable to store the name of the output file to store the complete set of utility patterns"""

        pass

    @abstractmethod
    def startMine(self):
        """Code for the mining process will start from this function"""

        pass

    @abstractmethod
    def getUtilityPatterns(self):
        """Complete set of utility patterns generated will be retrieved from this function"""

        pass

    @abstractmethod
    def storePatternsInFile(self, oFile):
        """Complete set of utility patterns will be saved in to an output file from this function

        :param oFile: Name of the output file
        :type oFile: file
        """

        pass

    @abstractmethod
    def getPatternsInDataFrame(self):
        """Complete set of utility patterns will be loaded in to data frame from this function"""

        pass

    @abstractmethod
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the program will be retrieved from this function"""

        pass

    @abstractmethod
    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the program will be retrieved from this function"""
        pass

    @abstractmethod
    def getRuntime(self):
        """Total amount of runtime taken by the program will be retrieved from this function"""

        pass



import sys
import functools
import pandas as pd
from PAMI.fuzzyFrequentSpatialPattern.basic.abstract import *


class FFList:
    """
     A class represent a Fuzzy List of an element

    Attributes :
    ----------
         item: int
             the item name
         sumiUtil: float
             the sum of utilities of an fuzzy item in database
         sumrUtil: float
             the sum of resting values of a fuzzy item in database
         elements: list
             a list of elements contain tid,Utility and resting values of element in each transaction
    Methods :
    -------
        addElement(element)
            Method to add an element to this fuzzy list and update the sums at the same time.

        printelement(e)
            Method to print elements            

    """

    def __init__(self, itemName):
        self.item = itemName
        self.sumiUtil = 0.0
        self.sumrUtil = 0.0
        self.elements = []

    def addElement(self, element):
        """
            A Method that add a new element to FFList

            :param element: an element to be add to FFList
            :pram type: Element
        """
        self.sumiUtil += element.iUtils
        self.sumrUtil += element.rUtils
        self.elements.append(element)

    def printelement(self):
        """
            A Method to Print elements in the FFList
        """
        for ele in self.elements:
            print(ele.tid, ele.iUtils, ele.rUtils)


class Element:
    """
        A class represents an Element of a fuzzy list

    Attributes :
    ----------
        tid : int
            keep tact of transaction id
        iUtils: float
            the utility of an fuzzy item in the transaction
        rUtils : float
            the nighbourhood resting value of an fuzzy item in the transaction
    """

    def __init__(self, tid, iUtil, rUtil):
        self.tid = tid
        self.iUtils = iUtil
        self.rUtils = rUtil


class Reagions:
    """
            A class calculate the regions

    Attributes :
    ----------
            low : int
                low region value
            middle: int 
                middle region value
            high : int
                high region values
        """

    def __init__(self, quantity, regionsNumber):
        self.low = 0
        self.middle = 0
        self.high = 0
        if regionsNumber == 3:  # if we have 3 regions
            if 0 < quantity <= 1:
                self.low = 1
                self.high = 0
                self.middle = 0
            elif 1 < quantity <= 6:
                self.low = float((6 - quantity) / 5)
                self.middle = float((quantity - 1) / 5)
                self.high = 0
            elif 6 < quantity <= 11:
                self.low = 0
                self.middle = float((11 - quantity) / 5)
                self.high = float((quantity - 6) / 5)
            else:
                self.low = 0
                self.middle = 0
                self.high = 1


class Pair:
    """
        A class to store item and it's quantity together
    """

    def __init__(self):
        self.item = 0
        self.quantity = 0


class ffsi(fuzzySpatialFrequentPatterns):
    """
        Fuzzy Frequent Spatial Pattern-Miner is desired to find all Spatially frequent fuzzy patterns
        which is on-trivial and challenging problem to its huge search space.we are using efficient pruning
         techniques to reduce the search space.

    Attributes :
    ----------
        iFile : file
            Name of the input file to mine complete set of fuzzy spatial frequent patterns
        oFile : file
               Name of the oFile file to store complete set of fuzzy spatial frequent patterns
        minSup : float
            The user given minimum support
        neighbors: map
            keep track of neighbours of elements
        memoryRSS : float
                To store the total amount of RSS memory consumed by the program
        startTime:float
               To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        itemsCnt: int
            To record the number of fuzzy spatial itemsets generated
        mapItemsLowSum: map
            To keep track of low region values of items
        mapItemsMidSum: map
            To keep track of middle region values of items
        mapItemsHighSum: map
            To keep track of high region values of items
        mapItemSum: map
            To keep track of sum of Fuzzy Vlues of items
        mapItemRegions: map
            To Kepp track of fuzzy regions of item
        jointCnt: int
            To keep track of the number of FFI-list that was constructed
        BufferSize: int
            represent the size of Buffer
        itemBuffer list
            to keep track of items in buffer
    Methods :
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function            
        convert(value):
            To convert the given user specified value
        FSFIMining( prefix, prefixLen, fsFim, minSup)
            Method generate FFI from prefix
        construct(px, py)
            A function to construct Fuzzy itemset from 2 fuzzy itemsets
        Intersection(nighb1,nighb2)
            Return common neighbours of 2 itemsset nighbours
        findElementWithTID(ulist, tid)
            To find element with same tid as given
        WriteOut(prefix, prefixLen, item, sumIutil,period)
            To Store the patten
    
    Executing the code on terminal :
    -------
        Format: python3 ffsi.py <inputFile> <outputFile> <neighbours> <minSup> <sep>
        Examples:  python3  ffsi.py sampleTDB.txt output.txt sampleN.txt 3  (minSup will be considered in support count or frequency)
                   python3  ffsi.py sampleTDB.txt output.txt sampleN.txt 0.3 (minSup and maxPer will be considered in percentage of database)
                                                            (will conseder "\t" as separator in both input and neighbourhood files)
                   python3  ffsi.py sampleTDB.txt output.txt sampleN.txt 3 , 
                                                              (will conseder "," as separator in both input and neighbourhood files)
    Sample run of importing the code:
    -------------------------------
        
        import ffsi as alg

        obj = alg.ffsi("input.txt", "neighbours.txt", 2)

        obj.startMine()

        fuzzySpatialFrequentPatterns = obj.getPatterns()

        print("Total number of fuzzy frequent spatial patterns:", len(fuzzySpatialFrequentPatterns))

        obj.storePatternsInFile("outp")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
            The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
    """
    startTime = float()
    endTime = float()
    minSup = str()
    maxPer = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    nFile =" "
    memoryUSS = float()
    memoryRSS = float()
    sep="\t"

    def __init__(self, iFile, nFile, minsup,sep="\t"):
        super().__init__(iFile, nFile, minsup,sep)
        self.mapItemNighbours = {}
        self.startTime = 0
        self.endTime = 0
        self.itemsCnt = 0
        self.mapItemsLowSum = {}
        self.mapItemsMidSum = {}
        self.mapItemsHighSum = {}
        self.mapItemSum = {}
        self.mapItemRegions = {}
        self.joinsCnt = 0
        self.BufferSize = 200
        self.itemsetBuffer = []
        self.finalPatterns = {}
        self.dbLen = 0

    def compareItems(self, o1, o2):
        """
            A Function that sort all FFI-list in asendng order of Support
        """
        compare = self.mapItemSum[o1.item] - self.mapItemSum[o2.item]
        if compare == 0:
            return o1.item - o2.item
        else:
            return compare

    def convert(self, value):
        """
        To convert the given user specified value
        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self.dbLen * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self.dbLen * value)
            else:
                value = int(value)
        return value

    def startMine(self):
        """ Frequent pattern mining process will start from here
        """
        self.startTime = time.time()
        with open(self.nFile, 'r') as file1:
            for line in file1:
                line=line.split("\n")[0]
                parts = line.split(self.sep)
                item = parts[0]
                neigh1 = []
                for i in range(1, len(parts)):
                    neigh1.append(parts[i])
                self.mapItemNighbours[item] = neigh1
        with open(self.iFile, 'r') as file:
            for line in file:
                line=line.split("\n")[0]
                parts = line.split(":")
                items = parts[0].split(self.sep)
                quanaities = parts[2].split(self.sep)
                self.dbLen += 1
                for i in range(0, len(items)):
                    regions = Reagions(int(quanaities[i]), 3)
                    item = items[i]
                    if item in self.mapItemsLowSum.keys():
                        low = self.mapItemsLowSum[item]
                        low += regions.low
                        self.mapItemsLowSum[item] = low
                    else:
                        self.mapItemsLowSum[item] = regions.low
                    if item in self.mapItemsMidSum.keys():
                        mid = self.mapItemsMidSum[item]
                        mid += regions.middle
                        self.mapItemsMidSum[item] = mid
                    else:
                        self.mapItemsMidSum[item] = regions.middle
                    if item in self.mapItemsHighSum.keys():
                        high = self.mapItemsHighSum[item]
                        high += regions.high
                        self.mapItemsHighSum[item] = high
                    else:
                        self.mapItemsHighSum[item] = regions.high
            listOfFFIlist = []
            mapItemsToFFLIST = {}
            self.minSup=self.convert(self.minSup)
            minSup = self.minSup
            print("minsup: ",minSup)
            for item1 in self.mapItemsLowSum.keys():
                item = item1
                low = self.mapItemsLowSum[item]
                mid = self.mapItemsMidSum[item]
                high = self.mapItemsHighSum[item]
                if low >= mid and low >= high:
                    self.mapItemSum[item] = low
                    self.mapItemRegions[item] = "L"
                elif mid >= low and mid >= high:
                    self.mapItemSum[item] = mid
                    self.mapItemRegions[item] = "M"
                elif high >= low and high >= mid:
                    self.mapItemRegions[item] = "H"
                    self.mapItemSum[item] = high
                if self.mapItemSum[item] >= self.minSup:
                    fuList = FFList(item)
                    mapItemsToFFLIST[item] = fuList
                    listOfFFIlist.append(fuList)
            listOfFFIlist.sort(key=functools.cmp_to_key(self.compareItems))
        tid = 0
        with open(self.iFile, 'r') as file:
            for line in file:
                line=line.split("\n")[0]
                parts = line.split(":")
                items = parts[0].split(self.sep)
                quanaities = parts[2].split(self.sep)
                revisedTransaction = []
                for i in range(0, len(items)):
                    pair = Pair()
                    pair.item = items[i]
                    regions = Reagions(int(quanaities[i]), 3)
                    item = pair.item
                    if self.mapItemSum[item] >= minSup:
                        if self.mapItemRegions[pair.item] == "L":
                            pair.quantity = regions.low
                        elif self.mapItemRegions[pair.item] == "M":
                            pair.quantity = regions.middle
                        elif self.mapItemRegions[pair.item] == "H":
                            pair.quantity = regions.high
                        if pair.quantity > 0:
                            revisedTransaction.append(pair)
                revisedTransaction.sort(key=functools.cmp_to_key(self.compareItems))
                for i in range(len(revisedTransaction) - 1, -1, -1):
                    pair = revisedTransaction[i]
                    remainUtil = 0
                    for j in range(len(revisedTransaction) - 1, i, -1):
                        if self.mapItemNighbours.get(pair.item) is None:
                            continue
                        if revisedTransaction[j].item in self.mapItemNighbours[pair.item]:
                            remainUtil += revisedTransaction[j].quantity
                    remaingUtility = remainUtil
                    if mapItemsToFFLIST.get(pair.item) is not None:
                        FFListOfItem = mapItemsToFFLIST[pair.item]
                        element = Element(tid, pair.quantity, remaingUtility)
                        FFListOfItem.addElement(element)
                tid += 1
        itemNeighbours = list(self.mapItemNighbours.keys())
        self.FSFIMining(self.itemsetBuffer, 0, listOfFFIlist, self.minSup, itemNeighbours)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def FSFIMining(self, prefix, prefixLen, FSFIM, minsup, itemNighbours):
        """Generates ffsi from prefix

        :param prefix: the prefix patterns of ffsi
        :type prefix: len
        :param prefixLen: the length of prefix
        :type prefixLen: int
           :param FSFIM: the Fuzzy list of prefix itemsets
           :type FSFIM: list
           :param minsup: the minimum support of 
           :type minsup:int
           :param itemNighbours: the set of common neighbours of prefix
           :type itemNighbours: set
        """
        for i in range(0, len(FSFIM)):
            X = FSFIM[i]
            if X.sumiUtil >= minsup:
                self.WriteOut(prefix, prefixLen, X.item, X.sumiUtil)
            newNeighbours = self.Intersection(self.mapItemNighbours.get(X.item), itemNighbours)
            if X.sumrUtil >= minsup:
                exULs = []
                for j in range(i + 1, len(FSFIM)):
                    Y = FSFIM[j]
                    if Y.item in newNeighbours:
                        exULs.append(self.construct(X, Y))
                        self.joinsCnt += 1
                self.itemsetBuffer.insert(prefixLen, X.item)
                self.FSFIMining(self.itemsetBuffer, prefixLen + 1, exULs, minsup, newNeighbours)

    def Intersection(self, nighb1, nighb2):
        """
            A function to get common neighbours from 2 itemsets
            :param nighb1: the set of neighbours of itemset 1
            :type nighb1: set
            :param nighb2: the set of neighbours of itemset 2
            :type nighb2: set
            :return : set of common neighbours of 2 itemsets
            :rtype :set
        """
        result = []
        if nighb1 is None or nighb2 is None:
            return result
        for i in range(0, len(nighb1)):
            if nighb1[i] in nighb2:
                result.append(nighb1[i])
        return result

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
       """
        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
       """
        return self.endTime - self.startTime

    def construct(self, px, py):
        """
            A function to construct a new Fuzzy itemset from 2 fuzzy itemsets

            :param px:the itemset px
            :type px:FFI-List
            :param py:ithemset py
            :type py:FFI-List
            :return :the itemset of pxy(px and py)
            :rtype :FFI-List
        """
        pxyUL = FFList(py.item)
        for ex in px.elements:
            ey = self.findElementWithTID(py, ex.tid)
            if ey is None:
                continue
            eXY = Element(ex.tid, min([ex.iUtils, ey.iUtils], key=lambda x: float(x)), ey.rUtils)
            pxyUL.addElement(eXY)
        return pxyUL

    def findElementWithTID(self, ulist, tid):
        """
            To find element with same tid as given
            :param ulist:fuzzylist
            :type ulist:FFI-List
            :param tid:transaction id
            :type tid:int
            :return:element eith tid as given
            :rtype: element if exizt or None
        """
        List = ulist.elements
        first = 0
        last = len(List) - 1
        while first <= last:
            mid = (first + last) >> 1
            if List[mid].tid < tid:
                first = mid + 1
            elif List[mid].tid > tid:
                last = mid - 1
            else:
                return List[mid]
        return None

    def WriteOut(self, prefix, prefixLen, item, sumIutil):
        """
            To Store the patten
            :param prefix: prefix of itemset
            :type prefix: list
            :param prefixLen: length of prefix
            :type prefixLen: int
            :param item: the last item
            :type item: int
            :param sumIutil: sum of utility of itemset
            :type sumIutil: float

        """
        self.itemsCnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i]) + "." + str(self.mapItemRegions[prefix[i]]) + " "
        res += str(item) + "." + str(self.mapItemRegions.get(item))
        res1 = str(sumIutil)
        self.finalPatterns[res] = res1

    # self.bwriter.write(res)

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = str(x) + " : " + str(y)
            writer.write("%s \n" % patternsAndSupport)


if __name__ == "__main__":
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6: # to  include a user specifed separator
            ap = ffsi(sys.argv[1], sys.argv[3], sys.argv[4],sys.argv[5])
        if len(sys.argv) == 5:  # to consider "\t" as a separator
            ap = ffsi(sys.argv[1], sys.argv[3], sys.argv[4])
        ap.startMine()
        fuzzySpatialFrequentPatterns = ap.getPatterns()
        print("Total number of fuzzy frequent Spatial Patterns:", len(fuzzySpatialFrequentPatterns))
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in seconds:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import sys
from PAMI.partialPeriodicFrequentPattern.basic.abstract import *


class PPF_DFS(partialPeriodicPatterns):
    """
    PPF_DFS is algorithm to mine the partial periodic frequent patterns.

    Attributes
    ----------

    path : file
        input file path
    output : file
        output file name
    minSup : float
        user defined minsup
    maxPer : float
        user defined maxPer
    minPR : float
        user defined minPR
    tidlist : dict
        it store tids each item
    last : int
        it represents last time stamp in database
    lno : int
        number of line in database
    mapSupport : dict
        to maintain the information of item and their frequency
    finalPatterns : dict
        it represents to store the patterns
    runTime : float
        storing the total runtime of the mining process
    memoryUSS : float
        storing the total amount of USS memory consumed by the program
    memoryRSS : float
        storing the total amount of RSS memory consumed by the program

    Methods
    -------
    findDelimiter(line)
        Identifying the delimiter of the input file
    getPer_Sup(tids)
        caluclate ip / (sup+1)
    getPerSup(tids)
        caluclate ip
    scanDatabase(path)
        scan all lines in database
    sava(prefix,suffix,tidsetx)
        save prefix pattern with support and periodic ratio
    Generation(prefix, itemsets, tidsets)
        Userd to implement prefix class equibalence method to generate the periodic patterns recursively
    startMine()
        Mining process will start from here
    getPartialPeriodicPatterns()
        Complete set of patterns will be retrieved with this function
    storePatterns InFile(ouputFile)
        Complete set of frequent patterns will be loaded in to a ouput file
    getPatternsInDataFrame()
        Complete set of frequent patterns will be loaded in to a ouput file
    getMemoryUSS()
        Total amount of USS memory consumed by the mining process will be retrieved from this function
    getMemoryRSS()
        Total amount of RSS memory consumed by the mining process will be retrieved from this function
    getRuntime()
        Total amount of runtime taken by the mining process will be retrieved from this function

    Format: python3 PPF_DFS.py <inputFile> <outputFile> <minSup> <maxPer> <minPR>
    Examples: python3 PPF_DFS.py sampleDB.txt patterns.txt 10 10 0.5
    """
    path = ' '
    iFile = ' '
    oFile = ' '
    minSup = float()
    maxPer = float()
    minPR = float()
    tidlist = {}
    last = 0
    lno = 0
    mapSupport = {}
    finalPatterns = {}
    runTime = float()
    memoryUSS = float()
    memoryRSS = float()
    startTime = float()
    endTime = float()

    def findDelimiter(self, line):
        """Identifying the delimiter of the input file
            :param line: list of special characters may be used by a user to split the items in a input file
            :type line: list of string
            :returns: Delimited string used in the input file to split each item
            :rtype: string
            """
        l = ['\t', ' ', '*', '&', ' ', '%', '$', '#', '@', '!', '    ', '*', '(', ')']
        j = None
        for i in l:
            if i in line:
                return i
        return j

    def getPer_Sup(self, tids):
        """
        calculate ip / (sup+1)
        :param tids: it represent tid list
        :type tids: list
        :return: ip / (sup+1)
        """
        # print(lno)
        tids = list(set(tids))
        tids.sort()
        per = 0
        sup = 0
        cur = 0
        if len(tids) == 0:
            return 0
        if abs(0 - tids[0]) <= self.maxPer:
            sup += 1
        for j in range(len(tids) - 1):
            i = j + 1
            per = abs(tids[i] - tids[j])
            if (per <= self.maxPer):
                sup += 1
            cur = tids[j]
        if abs(self.last - tids[len(tids) - 1]) <= self.maxPer:
            sup += 1
        if sup == 0:
            return 0
        return sup / (len(tids) + 1)

    def getPerSup(self, tids):
        """
        calculate ip of a pattern
        :param tids: tid list of the pattern
        :type tids: list
        :return: ip
        """
        # print(lno)
        tids = list(set(tids))
        tids.sort()
        per = 0
        sup = 0
        cur = 0
        if len(tids) == 0:
            return 0
        if abs(0 - tids[0]) <= self.maxPer:
            sup += 1
        for j in range(len(tids) - 1):
            i = j + 1
            per = abs(tids[i] - tids[j])
            if (per <= self.maxPer):
                sup += 1
        if abs(tids[len(tids) - 1] - self.last) <= self.maxPer:
            sup += 1
        if sup == 0:
            return 0
        return sup

    def scanDatabase(self, path):
        """
        scan all lines of database and create support list
        :param path: it represents input file name
        :return: support list each item
        """
        id1 = 0
        with open(path, 'r') as f:
            for line in f:
                self.lno += 1
                # first=min(first,lno)
                # last=max(last,lno)
                line = line.strip()
                delimiter = self.findDelimiter([*line])
                s = [i.rstrip() for i in line.split(delimiter)]
                n = int(s[0])
                self.last = max(self.last, n)
                for i in range(1, len(s)):
                    si = s[i]
                    if abs(0 - n) <= self.maxPer:
                        if si not in self.mapSupport:
                            self.mapSupport[si] = [1, 1, n]
                            self.tidlist[si] = [n]
                        else:
                            lp = abs(n - self.mapSupport[si][2])
                            if lp <= self.maxPer:
                                self.mapSupport[si][0] += 1
                            self.mapSupport[si][1] += 1
                            self.mapSupport[si][2] = n
                            self.tidlist[si].append(n)
                    else:
                        if si not in self.mapSupport:
                            self.mapSupport[si] = [0, 1, n]
                            self.tidlist[si] = [n]
                        else:
                            lp = abs(n - self.mapSupport[si][2])
                            if lp <= self.maxPer:
                                self.mapSupport[si][0] += 1
                            self.mapSupport[si][1] += 1
                            self.mapSupport[si][2] = n
                            self.tidlist[si].append(n)
        for x, y in self.mapSupport.items():
            lp = abs(self.last - self.mapSupport[x][2])
            if lp <= self.maxPer:
                self.mapSupport[x][0] += 1
        self.mapSupport = {k: [v[1], v[0]] for k, v in self.mapSupport.items() if
                           v[1] >= self.minSup and v[0] / (self.minSup + 1) >= self.minPR}
        plist = [key for key, value in sorted(self.mapSupport.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        return plist

    def save(self, prefix, suffix, tidsetx):
        """
        sava prefix patterns with support and periodic ratio
        :param prefix: prefix patterns
        :type prefix: list
        :param suffix: it represents suffix itemsets
        :type suffix: list
        :param tidsetx: it represents prefix tids
        :type tidsetx: list
        """
        tidsetx = list(set(tidsetx))
        if (prefix == None):
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = self.getPerSup(tidsetx)
        val1 = self.getPer_Sup(tidsetx)
        # print(prefix,tidsetx,val,val1)
        if len(tidsetx) >= self.minSup and val / (len(tidsetx) + 1) >= self.minPR:
            """self.itemsetCount+=1
            s1=str(prefix)+":"+str(len(tidsetx))+":"+str(val1)
            self.writer.write('%s \n'%s1)"""
            self.finalPatterns[tuple(prefix)] = [len(tidsetx), val1]

    def Generation(self, prefix, itemsets, tidsets):
        """
        here equibalence class is followed amd checks fro the patterns generated for periodic frequent patterns.
        :param prefix: main equivalence prefix
        :type prefix: periodic-frequent item or pattern
        :param itemsets: patterns which are items combined with prefix and satisfythe periodicity
                        and frequent wthi their time stamps
        :type itemsets: list
        :param tidsets: time stamps of the items in the argument itemSets
        :type tidsets: list
        """
        if (len(itemsets) == 1):
            i = itemsets[0]
            tidi = tidsets[0]
            self.save(prefix, [i], tidi)
            return
        for i in range(len(itemsets)):
            itemx = itemsets[i]
            if (itemx == None):
                continue
            tidsetx = tidsets[i]
            classItemsets = []
            classtidsets = []
            itemsetx = [itemx]
            for j in range(i + 1, len(itemsets)):
                itemj = itemsets[j]
                tidsetj = tidsets[j]
                y = list(set(tidsetx) & set(tidsetj))
                val = self.getPerSup(y)
                # if(len(y)>=minsup and val/(len(y)+1)>=minpr):
                if len(y) >= self.minSup and val / (self.minSup + 1) >= self.minPR:
                    classItemsets.append(itemj)
                    classtidsets.append(y)
            newprefix = list(set(itemsetx)) + prefix
            self.Generation(newprefix, classItemsets, classtidsets)
            self.save(prefix, list(set(itemsetx)), tidsetx)

    def startMine(self):
        """
        Main program start with extracting the periodic frequent items from the database and
        performs prefix equivalence to form the combinations and generates closed periodic frequent patterns.
        """
        self.path = self.iFile
        starttime = time.time()
        plist = self.scanDatabase(self.path)
        print(len(plist))
        for i in range(len(plist)):
            itemx = plist[i]
            tidsetx = self.tidlist[itemx]
            itemsetx = [itemx]
            itemsets = []
            tidsets = []
            for j in range(i + 1, len(plist)):
                itemj = plist[j]
                tidsetj = self.tidlist[itemj]
                y1 = list(set(tidsetx) & set(tidsetj))
                val = self.getPerSup(y1)
                # if(len(y1)>=minsup and val/(len(y1)+1)>=minpr):
                if len(y1) >= self.minSup and val / (self.minSup + 1) >= self.minPR:
                    itemsets.append(itemj)
                    tidsets.append(y1)
            self.Generation(itemsetx, itemsets, tidsets)
            self.save(None, itemsetx, tidsetx)
        # print("eclat Total Itemsets:",self.itemsetCount)
        endtime = time.time()
        self.runTime = (endtime - starttime)
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        # print("eclat Time taken:",temp)
        # print("eclat Memory Space:",resource.getrusage(resource.RUSAGE_SELF).ru_maxrss)

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.runTime

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataframe = pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataframe

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            s1 = str(x) + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPartialPeriodicPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == '__main__':
    if len(sys.argv) == 6:
        ap = PPF_DFS(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.startMine()
        frequentPatterns = ap.getPartialPeriodicPatterns()
        print(f"Total number of Frequent Patterns: {len(frequentPatterns)}")
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print(f'Total Memory in USS: {memUSS}')
        memRSS = ap.getMemoryRSS()
        print(f'Total Memory in RSS: {memRSS}')
        run = ap.getRuntime()
        print(f'Total ExecutionTime in seconds: {run}')
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")





#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.


import sys
from PAMI.frequentPattern.closed.abstract import *


class Charm(frequentPatterns):
    """ CHARM is one of the fundamental algorithm to discover closed frequent patterns in a transactional database.
        Closed frequent patterns are patterns if there exists no superset that has the same support count as this original itemset.
        This program employs downward closure property to  reduce the search space effectively.
        This algorithm employs depth-first search technique to find the complete set of closed frequent patterns in a
        transactional database.
        
        Reference:
        ----------
            Mohammed J. Zaki and Ching-Jui Hsiao, CHARM: An Efficient Algorithm for Closed Itemset Mining,
            Proceedings of the 2002 SIAM, SDM. 2002, 457-473, https://doi.org/10.1137/1.9781611972726.27

       Attributes
       ----------
            iFile : file
                Name of the Input file or path of the input file
            oFile : file
                Name of the output file or path of the output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            startTime: float
                To record the start time of the mining process
            endTime: float
                To record the completion time of the mining process
            Database : list
                To store the transactions of a database in list
            mapSupport : Dictionary
                To maintain the information of item and their frequency
            lno : int
                it represents the total no of transactions
            tree : class
                it represents the Tree class
            itemSetCount : int
                it represents the total no of patterns
            finalPatterns : dict
                it represents to store the patterns
            tidList : dict
                stores the timestamps of an item
            hashing : dict
                stores the patterns with their support to check for the closed property

        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            storePatternsInFile(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsInDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            creatingItemsets()
                Stores the frequent patterns with their timestamps from the dataset
        
        
        Executing the code on terminal:
        -------
            Format: python3 charm.py <inputFile> <outputFile> <minSup>

            Examples:

            python3 charm.py sampleDB.txt patterns.txt 10.0   (minSup will be considered in percentage of database transactions)

            python3 charm.py sampleDB.txt patterns.txt 10     (minSup will be considered in support count or frequency)

        Sample run of the importing code:
        --------------

            from PAMI.frequentPattern.closed import closed as alg

            obj = alg.Closed(iFile, minSup)

            obj.startMine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.storePatternsInFile(oFile)

            Df = obj.getPatternsInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

        """

    startTime = float()
    endTime = float()
    minSup = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    sep = " "
    memoryUSS = float()
    memoryRSS = float()
    Database = []
    tidList = {}
    lno = 0
    mapSupport = {}
    hashing = {}
    itemSetCount = 0
    maxItemId = 0
    tableSize = 10000
    writer = None

    def convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self.lno * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self.lno * value)
            else:
                value = int(value)
        return value

    def creatingItemsets(self):
        """
        Storing the complete frequent patterns of the database/input file in a database variable
        """
        with open(self.iFile, 'r') as f:
            for line in f:
                i = [i.rstrip() for i in line.split(self.sep)]
                self.lno += 1
                n = self.lno
                for j in i:
                    if j not in self.mapSupport:
                        self.mapSupport[j] = 1
                        self.tidList[j] = [n]
                    else:
                        self.mapSupport[j] += 1
                        self.tidList[j].append(n)
        self.minSup = self.convert(self.minSup)
        self.mapSupport = {k: v for k, v in self.mapSupport.items() if v >= self.minSup}
        flist = {}
        self.tidList = {k: v for k, v in self.tidList.items() if k in self.mapSupport}
        for x, y in self.tidList.items():
            t1 = 0
            for i in y:
                t1 += i
            flist[x] = t1
        flist = [key for key, value in sorted(flist.items(), key=lambda x: x[1])]
        return flist

    def calculate(self, tidSet):
        """To calculate the hashcode of pattern

            :param tidSet: the timestamps of a pattern

            :type tidSet: list

            :rtype: int
        """

        hashcode = 0
        for i in tidSet:
            hashcode += i
        if hashcode < 0:
            hashcode = abs(0 - hashcode)
        return hashcode % self.tableSize

    def contains(self, itemSet, value, hashcode):
        """ Check for the closed property(patterns with same support) by checking the hashcode(sum of timestamps),
            if hashcode key in hashing dict is none then returns a false, else returns with true.

            :param itemSet: frequent pattern

            :type itemSet: list

            :param value: support of the pattern

            :type value: int

            :param hashcode: calculated from the timestamps of pattern

            :type hashcode: int
            """
        if self.hashing.get(hashcode) is None:
            return False
        for i in self.hashing[hashcode]:
            itemSetx = i
            if value == self.hashing[hashcode][itemSetx] and set(itemSetx).issuperset(itemSet):
                return True
        return False

    def save(self, prefix, suffix, tidSetx):
        """ Check for the closed property (patterns with same support), if found deletes the subsets and stores
            supersets and also saves the patterns that satisfy the closed property

            :param prefix: the prefix of a pattern

            :param suffix: the suffix of a patterns

            :type suffix: list

            :param tidSetx: the timestamp of a patterns

            :type tidSetx: list
        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        prefix = list(set(prefix))
        prefix.sort()
        val = len(tidSetx)
        if val >= self.minSup:
            hashcode = self.calculate(tidSetx)
            if self.contains(prefix, val, hashcode) is False:
                sample = str()
                for i in prefix:
                    sample = sample + i + " "
                self.itemSetCount += 1
                self.finalPatterns[sample] = val
            if hashcode not in self.hashing:
                self.hashing[hashcode] = {tuple(prefix): val}
            else:
                self.hashing[hashcode][tuple(prefix)] = val

    def processEquivalenceClass(self, prefix, itemSets, tidSets):
        """ Equivalence class is followed  and check for the patterns which satisfies frequent properties.

            :param prefix:  main equivalence prefix

            :type prefix: frequent item or pattern

            :param itemSets: patterns which are items combined with prefix and satisfying the minSup

            :type itemSets: list

            :param tidSets: timestamps of the items in the argument itemSets

            :type tidSets: list


        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidI = tidSets[0]
            self.save(prefix, [i], tidI)
            return
        if len(itemSets) == 2:
            itemX = itemSets[0]
            tidSetX = tidSets[0]
            itemY = itemSets[1]
            tidSetY = tidSets[1]
            y1 = list(set(tidSetX).intersection(tidSetY))
            if len(y1) >= self.minSup:
                suffix = []
                suffix += [itemX, itemY]
                suffix = list(set(suffix))
                self.save(prefix, suffix, y1)
            if len(y1) != len(tidSetX):
                self.save(prefix, [itemX], tidSetX)
            if len(y1) != len(tidSetY):
                self.save(prefix, [itemX], tidSetY)
            return
        for i in range(len(itemSets)):
            itemX = itemSets[i]
            if itemX is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetx = [itemX]
            for j in range(i + 1, len(itemSets)):
                itemY = itemSets[j]
                if itemY is None:
                    continue
                tidSetY = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetY))
                if len(y) < self.minSup:
                    continue
                if len(tidSetX) == len(tidSetY) and len(y) == len(tidSetX):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    itemSetx.append(itemY)
                elif len(tidSetX) < len(tidSetY) and len(y) == len(tidSetX):
                    itemSetx.append(itemY)
                elif len(tidSetX) > len(tidSetY) and len(y) == len(tidSetY):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    classItemSets.append(itemY)
                    classTidSets.append(y)
                else:
                    classItemSets.append(itemY)
                    classTidSets.append(y)
            if len(classItemSets) > 0:
                newPrefix = list(set(itemSetx)) + prefix
                self.processEquivalenceClass(newPrefix, classItemSets, classTidSets)
                self.save(prefix, list(set(itemSetx)), tidSetX)

    def startMine(self):
        """
        Mining process will start from here by extracting the frequent patterns from the database. It performs prefix
        equivalence to generate the combinations and generates closed frequent patterns.
        """
        self.startTime = time.time()
        plist = self.creatingItemsets()
        for i in range(len(plist)):
            itemX = plist[i]
            if itemX is None:
                continue
            tidSetx = self.tidList[itemX]
            itemSetx = [itemX]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(plist)):
                itemY = plist[j]
                if itemY is None:
                    continue
                tidSetY = self.tidList[itemY]
                y1 = list(set(tidSetx).intersection(tidSetY))
                if len(y1) < self.minSup:
                    continue
                if len(tidSetx) == len(tidSetY) and len(y1) == len(tidSetx):
                    plist.insert(j, None)
                    itemSetx.append(itemY)
                elif len(tidSetx) < len(tidSetY) and len(y1) == len(tidSetx):
                    itemSetx.append(itemY)
                elif len(tidSetx) > len(tidSetY) and len(y1) == len(tidSetY):
                    plist.insert(j, None)
                    itemSets.append(itemY)
                    tidSets.append(y1)
                else:
                    itemSets.append(itemY)
                    tidSets.append(y1)
            if len(itemSets) > 0:
                self.processEquivalenceClass(itemSetx, itemSets, tidSets)
            self.save(None, itemSetx, tidSetx)
        print("Closed Frequent patterns were generated successfully using charm algorithm")
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataframe = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """

        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 4 or len(sys.argv) == 5:
        if len(sys.argv) == 5:
            ap = Charm(sys.argv[1], sys.argv[3], sys.argv[4])
        if len(sys.argv) == 4:
            ap = Charm(sys.argv[1], sys.argv[3])
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Closed Frequent Patterns:", len(Patterns))
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
from PAMI.uncertainPeriodicFrequentPattern.basic.abstract import *

minSup = float()
maxPer = float()
lno = int()
first = int()
last = int()
periodic = {}


class Item:
    """
                A class used to represent the item with probability in transaction of dataset

                ...

                Attributes
                __________
                item : int or word
                    Represents the name of the item
                probability : float
                    Represent the existential probability(likelihood presence) of an item
            """

    def __init__(self, item, probability):
        self.item = item
        self.probability = probability


class Node(object):
    """
            A class used to represent the node of frequentPatternTree

                ...

                Attributes
                ----------
                item : int
                    storing item of a node
                probability : int
                    To maintain the expected support of node
                parent : node
                    To maintain the parent of every node
                children : list
                    To maintain the children of node

                Methods
                -------

                addChild(itemName)
                    storing the children to their respective parent nodes
            """

    def __init__(self, item, children):
        self.item = item
        self.probability = 1
        self.secondProbability = 1
        self.p = 1
        self.children = children
        self.parent = None
        self.TimeStamps = []

    def addChild(self, node):
        """
            to add children details to parent node
            :param node: children node
            :return: update parent node children
        """
        self.children[node.item] = node
        node.parent = self


class Tree(object):
    """
                A class used to represent the frequentPatternGrowth tree structure

                ...

                Attributes
                ----------
                root : Node
                    Represents the root node of the tree
                summaries : dictionary
                    storing the nodes with same item name
                info : dictionary
                    stores the support of items


                Methods
                -------
                addTransaction(transaction)
                    creating transaction as a branch in frequentPatternTree
                addConditionnalTransaction(prefixPaths, supportOfItems)
                    construct the conditional tree for prefix paths
                conditionalPatterns(Node)
                    generates the conditional patterns from tree for specific node
                condtionaltrasnactions(prefixPaths,Support)
                    takes the prefixPath of a node and support at child of the path and extract the frequent items from
                    prefixPaths and generates prefixPaths with items which are frequent
                remove(Node)
                    removes the node from tree once after generating all the patterns respective to the node
                generatePatterns(Node)
                    starts from the root node of the tree and mines the frequent patterns

            """

    def __init__(self):
        self.root = Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, tid):
        """adding transaction into tree

                                :param transaction : it represents the one transactions in database
                                :type transaction : list
                                :param tid : the timestamp of transaction
                                :type tid : list
                                """
        currentNode = self.root
        k = 0
        for i in range(len(transaction)):
            k += 1
            if transaction[i].item not in currentNode.children:
                newNode = Node(transaction[i].item, {})
                newNode.k = k
                newNode.secondProbability = transaction[i].probability
                l1 = i - 1
                temp = []
                while l1 >= 0:
                    temp.append(transaction[l1].probability)
                    l1 -= 1
                if len(temp) == 0:
                    newNode.probability = round(transaction[i].probability, 2)
                else:
                    newNode.probability = round(max(temp) * transaction[i].probability, 2)
                currentNode.addChild(newNode)
                if transaction[i].item in self.summaries:
                    self.summaries[transaction[i].item].append(newNode)
                else:
                    self.summaries[transaction[i].item] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i].item]
                currentNode.secondProbability = max(transaction[i].probability, currentNode.secondProbability)
                currentNode.k = k
                l1 = i - 1
                temp = []
                while l1 >= 0:
                    temp.append(transaction[l1].probability)
                    l1 -= 1
                if len(temp) == 0:
                    currentNode.probability += round(transaction[i].probability, 2)
                else:
                    nn = max(temp) * transaction[i].probability
                    currentNode.probability += round(nn, 2)
        currentNode.TimeStamps = currentNode.TimeStamps + tid

    def addConditionalPatterns(self, transaction, tid, sup):
        """constructing conditional tree from prefixPaths

                :param transaction : it represents the one transactions in database
                :type transaction : list
                :param tid : timestamps of a pattern or transaction in tree
                :param tid : list
                :param sup : support of prefixPath taken at last child of the path
                :type sup : int
                                        """
        currentNode = self.root
        k = 0
        for i in range(len(transaction)):
            k += 1
            if transaction[i] not in currentNode.children:
                newNode = Node(transaction[i], {})
                newNode.k = k
                newNode.probability = sup
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
                currentNode.k = k
                currentNode.probability += sup
        currentNode.TimeStamps = currentNode.TimeStamps + tid

    def conditionalPatterns(self, alpha):
        """generates all the conditional patterns of respective node

                :param alpha : it represents the Node in tree
                :type alpha : Node
                                """
        finalPatterns = []
        finalSets = []
        sup = []
        for i in self.summaries[alpha]:
            set1 = i.TimeStamps
            s = i.probability
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
                sup.append(s)
        finalPatterns, finalSets, support, info = self.conditionalTransactions(finalPatterns, finalSets, sup)
        return finalPatterns, finalSets, support, info

    def removeNode(self, nodeValue):
        """removing the node from tree

                :param nodeValue : it represents the node in tree
                :type nodeValue : node
                               """
        for i in self.summaries[nodeValue]:
            i.parent.TimeStamps = i.parent.TimeStamps + i.TimeStamps
            del i.parent.children[nodeValue]

    def get_ts(self, alpha):
        temp = []
        for i in self.summaries[alpha]:
            temp += i.TimeStamps
        return temp

    def getPeriodAndSupport(self, support, TimeStamps):
        """

        Parameters
        ----------
        support
        TimeStamps

        Returns
        -------
        support and period

        """
        global maxPer
        global lno
        TimeStamps.sort()
        cur = 0
        per = 0
        sup = support
        for j in range(len(TimeStamps)):
            per = max(per, TimeStamps[j] - cur)
            if per > maxPer:
                return [0, 0]
            cur = TimeStamps[j]
            # sup+=1
        per = max(per, lno - cur)
        return [sup, per]

    def conditionalTransactions(self, conditionalPatterns, conditionalTimeStamps, support):
        """ It generates the conditional patterns with frequent items

            :param conditionalPatterns : conditional patterns generated from conditionalPatterns() method for respective node
            :type conditionalPatterns : list
            :param conditionalTimeStamps : timestamps of respective conditional timestamps
            :type conditionalTimeStamps : list
            :param support : the support of conditional pattern in tree
            :type support : list
            """
        global minSup, maxPer, lno
        pat = []
        TimeStamps = []
        sup = []
        data1 = {}
        count = {}
        for i in range(len(conditionalPatterns)):
            for j in conditionalPatterns[i]:
                if j in data1:
                    data1[j] = data1[j] + conditionalTimeStamps[i]
                    count[j] += support[i]
                else:
                    data1[j] = conditionalTimeStamps[i]
                    count[j] = support[i]
        updatedDict = {}
        for m in data1:
            updatedDict[m] = self.getPeriodAndSupport(count[m], data1[m])
        updatedDict = {k: v for k, v in updatedDict.items() if v[0] >= minSup and v[1] <= maxPer}
        count = 0
        for p in conditionalPatterns:
            p1 = [v for v in p if v in updatedDict]
            trans = sorted(p1, key=lambda x: (updatedDict.get(x)[0]), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                TimeStamps.append(conditionalTimeStamps[count])
                sup.append(support[count])
            count += 1
        return pat, TimeStamps, sup, updatedDict

    def generatePatterns(self, prefix):
        """generates the patterns

                                :param prefix : forms the combination of items
                                :type prefix : list
                                """
        global periodic, minSup
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x)[0])):
            pattern = prefix[:]
            pattern.append(i)
            s = 0
            for x in self.summaries[i]:
                if x.k <= 2:
                    s += x.probability
                elif x.k >= 3:
                    n = x.probability * pow(x.secondProbability, (x.k - 2))
                    s += n
            periodic[tuple(pattern)] = self.info[i]
            if s >= minSup:
                patterns, TimeStamps, support, info = self.conditionalPatterns(i)
                conditionalTree = Tree()
                conditionalTree.info = info.copy()
                for pat in range(len(patterns)):
                    conditionalTree.addConditionalPatterns(patterns[pat], TimeStamps[pat], support[pat])
                if len(patterns) > 0:
                    conditionalTree.generatePatterns(pattern)
            self.removeNode(i)


class PTubeP(periodicFrequentPatterns):
    """
        Periodic-TubeP is  to discover periodic-frequent patterns in a temporal database.

        Reference:
        --------

       Attributes
        ----------
        iFile : file
            Name of the Input file to mine complete set of periodic-frequent patterns
        oFile : file
            Name of the output file to store complete set of periodic-frequent patterns
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        minSup : int/float
            The user given minimum support
        maxPer : int/float
            The user given maximum period
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            To represent the total no of transaction
        tree : class
            To represents the Tree class
        itemSetCount : int
            To represents the total no of patterns
        finalPatterns : dict
            To store the complete patterns
        Methods
        -------
        startMine()
            Mining process will start from here
        getPeriodicFrequentPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets(fileName)
            Scans the dataset and stores in a list format
        PeriodicFrequentOneItem()
            Extracts the one-periodic-frequent patterns from database
        updateDatabases()
            Update the database by removing aperiodic items and sort the Database by item decreased support
        buildTree()
            After updating the Database, remaining items will be added into the tree by setting root node as null
        convert()
            to convert the user specified value
        startMine()
            Mining process will start from this function
        Executing the code on terminal:
        -------
        Format:
        ------
        python3 PTubeP.py <inputFile> <outputFile> <minSup> <maxPer>
        Examples:
        --------
        python3 PTubeP.py sampleTDB.txt patterns.txt 0.3 4     (minSup and maxPer will be considered in support count or frequency)
        Sample run of importing the code:
        -------------------

        from PAMI.uncertainPeriodicFrequentPattern.basic import PTubeP as alg

        obj = alg.PTubeP(iFile, minSup, maxPer)

        obj.startMine()

        periodicFrequentPatterns = obj.getPeriodicFrequentPatterns()

        print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

        obj.storePatternsInFile(oFile)

        Df = obj.getPatternsInDataFrame()

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

    """
    startTime = float()
    endTime = float()
    minSup = float()
    maxPer = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    memoryUSS = float()
    memoryRSS = float()
    Database = []
    rank = {}
    lno = 0

    def creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        try:
            with open(self.iFile, 'r') as f:
                for line in f:
                    temp = [i.rstrip() for i in line.split("\t")]
                    tr = [int(temp[0])]
                    for i in temp[1:]:
                        i1 = i.index('(')
                        i2 = i.index(')')
                        item = i[0:i1]
                        probability = float(i[i1 + 1:i2])
                        product = Item(item, probability)
                        tr.append(product)
                    self.lno += 1
                    self.Database.append(tr)
        except IOError:
            print("File Not Found")

    def scanDatabase(self):
        """takes the transactions and calculates the support of each item in the dataset and assign the
                            ranks to the items by decreasing support and returns the frequent items list

        """
        global first, last
        mapSupport = {}
        for i in self.Database:
            n = int(i[0])
            for j in i[1:]:
                if j.item not in mapSupport:
                    mapSupport[j.item] = [round(j.probability, 2), abs(first - n), n]
                else:
                    mapSupport[j.item][0] += round(j.probability, 2)
                    mapSupport[j.item][1] = max(mapSupport[j.item][1], abs(n - mapSupport[j.item][2]))
                    mapSupport[j.item][2] = n
        for key in mapSupport:
            mapSupport[key][1] = max(mapSupport[key][1], last - mapSupport[key][2])
        mapSupport = {k: [round(v[0], 2), v[1]] for k, v in mapSupport.items() if
                      v[1] <= self.maxPer and v[0] >= self.minSup}
        plist = [k for k, v in sorted(mapSupport.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        self.rank = dict([(index, item) for (item, index) in enumerate(plist)])
        return mapSupport, plist

    def buildTree(self, data, info):
        """it takes the transactions and support of each item and construct the main tree with setting root
                            node as null

                                :param data : it represents the one transactions in database
                                :type data : list
                                :param info : it represents the support of each item
                                :type info : dictionary
                                """
        rootNode = Tree()
        rootNode.info = info.copy()
        for i in range(len(data)):
            set1 = [data[i][0]]
            rootNode.addTransaction(data[i][1:], set1)
        return rootNode

    def updateTransactions(self, dict1):
        """remove the items which are not frequent from transactions and updates the transactions with rank of items

                :param dict1 : frequent items with support
                :type dict1 : dictionary
                """
        list1 = []
        for tr in self.Database:
            list2 = [int(tr[0])]
            for i in range(1, len(tr)):
                if tr[i].item in dict1:
                    list2.append(tr[i])
            if len(list2) >= 2:
                basket = list2[1:]
                basket.sort(key=lambda val: self.rank[val.item])
                list2[1:] = basket[0:]
                list1.append(list2)
        return list1

    def Check(self, i, x):
        """To check the presence of item or pattern in transaction

                                :param x: it represents the pattern
                                :type x : list
                                :param i : represents the uncertain transactions
                                :type i : list
                                """
        for m in x:
            k = 0
            for n in i:
                if m == n.item:
                    k += 1
            if k == 0:
                return 0
        return 1

    def convert(self, value):
        """
            To convert the given user specified value
            :param value: user specified value
            :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = float(value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
            else:
                value = int(value)

        return value

    def removeFalsePositives(self):
        """
        To remove false positives in generated patterns
        :return: original patterns
        """
        periods = {}
        for i in self.Database:
            for x, y in periodic.items():
                if len(x) == 1:
                    periods[x] = y[0]
                else:
                    s = 1
                    check = self.Check(i[1:], x)
                    if check == 1:
                        for j in i[1:]:
                            if j.item in x:
                                s *= j.probability
                        if x in periods:
                            periods[x] += s
                        else:
                            periods[x] = s
        for x, y in periods.items():
            if y >= minSup:
                sample = str()
                for i in x:
                    sample = sample + i + " "
                self.finalPatterns[sample] = y

    def startMine(self):
        """Main method where the patterns are mined by constructing tree and remove the remove the false patterns
                           by counting the original support of a patterns


               """
        global minSup, maxPer, first, last, lno
        self.startTime = time.time()
        self.creatingItemSets()
        self.minSup = self.convert(self.minSup)
        self.maxPer = self.convert(self.maxPer)
        minSup, maxPer, lno = self.minSup, self.maxPer, len(self.Database)
        mapSupport, plist = self.scanDatabase()
        updatedTrans = self.updateTransactions(mapSupport)
        info = {k: v for k, v in mapSupport.items()}
        root = self.buildTree(updatedTrans, info)
        root.generatePatterns([])
        self.removeFalsePositives()
        print("Periodic Frequent patterns were generated successfully using Periodic-TubeP algorithm")
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataframe = pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataframe

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPeriodicFrequentPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    if len(sys.argv) is 5:
        ap = PTubeP(sys.argv[1], sys.argv[3], sys.argv[5])
        ap.startMine()
        frequentPatterns = ap.getPeriodicFrequentPatterns()
        print("Total number of Frequent Patterns:", len(frequentPatterns))
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in seconds:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import sys
import time
from PAMI.localPeriodicPattern.basic.abstract import *

class LPPMBreadth(localPeriodicPatterns):
    """

        Attributes:
        -----------
        iFile : str
            Input file name or path of the input file
        oFile : str
            Output file name or path of the output file
        maxPer : float
            User defined maxPer value.
        maxSoPer : float
            User defined maxSoPer value.
        minDur : float
            User defined minDur value.
        tsmin : int / date
            First time stamp of input data.
        tsmax : int / date
            Last time stamp of input data.
        startTime : float
            Time when start of execution the algorithm.
        endTime : float
            Time when end of execution the algorithm.
        finalPatterns : dict
            To store local periodic patterns and its PTL.
        tsList : dict
            To store items and its time stamp as bit vector.
        :param sep: separator used to distinguish items from each other. The default separator is tab space.
        :type sep: str

        Methods
        -------
        findSeparator(line)
            Find the separator of the line which split strings.
        createTSlist()
            Create the TSlist as bit vector from input data.
        generateLPP()
            Generate 1 length local periodic pattens by TSlist and execute depth first search.
        calculatePTL(tsList)
            Calculate PTL from input tsList as bit vector
        LPPMBreathSearch(extensionOfP)
            Mining local periodic patterns using breadth first search.
        startMine()
            Mining process will start from here.
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function.
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function.
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function.
        getLocalPeriodicPatterns()
            return local periodic patterns and its PTL
        storePatternsInFile(oFile)
            Complete set of local periodic patterns will be loaded in to a output file.
        getPatternsInDataFrame()
            Complete set of local periodic patterns will be loaded in to a dataframe.

        Executing the code on terminal
        ------------------------------
        Format: python3 LPPBreadth.py <inputFile> <outputFile> <maxPer> <minSoPer> <minDur>
        Examples: python3 LPPMBreadth.py sampleDB.txt patterns.txt 0.3 0.4 0.5

        Sample run of importing the code
        --------------------------------
        from PAMI.localPeriodicPattern.basic import LPPMBreadth as alg
        obj = alg.LPPMBreadth(iFile, maxPer, maxSoPer, minDur)
        obj.startMine()
        localPeriodicPatterns = obj.getLocalPeriodicPatterns()
        print(f'Total number of local periodic patterns: {len(localPeriodicPatterns)}')
        obj.storePatternsInFile(oFile)
        Df = obj.getPatternsInDataFrame()
        memUSS = obj.getMemoryUSS()
        print(f'Total memory in USS: {memUSS}')
        memRSS = obj.getMemoryRSS()
        print(f'Total memory in RSS: {memRSS}')
        runtime = obj.getRuntime()
        print(f'Total execution time in seconds: {runtime})

        Credits
        -------
        The complete program was written by So Nakamura under the supervision of Professor Rage Uday Kiran.
    """
    iFile = ' '
    oFile = ' '
    maxPer = float()
    maxSoPer = float()
    minDur = float()
    tsmin = 0
    tsmax = 0
    startTime = float()
    endTime = float()
    memoryUSS = float()
    memoryRSS = float()
    finalPatterns = {}
    tsList = {}
    sep = ' '

    def findDelimiter(self, line):
        """Identifying the delimiter of the input file
            :param line: list of special characters may be used by a user to split the items in a input file
            :type line: list of string
            :returns: Delimited string used in the input file to split each item
            :rtype: string
            """
        l = ['\t', ',', '*', '&', ' ', '%', '$', '#', '@', '!', '    ', '*', '(', ')']
        j = None
        for i in l:
            if i in line:
                return i
        return j

    def createTSlist(self):
        """
        Create tsList as bit vector from temporal data.
        """
        with open(self.iFile, 'r') as f:
            count = 1
            bitVector = 0b1 << count
            bitVector = bitVector | 0b1
            line = f.readline()
            line = line.strip()
            separator = self.findDelimiter(line)
            # line = [item for item in line.split(separator)]
            line = [item for item in line.split(self.sep)]
            self.tsmin = int(line.pop(0))
            self.tsList = {item: bitVector for item in line}
            count += 1
            for line in f:
                bitVector = 0b1 << count
                bitVector = bitVector | 0b1
                line = line.strip()
                # line = [item for item in line.split(separator)]
                line = [item for item in line.split(self.sep)]
                ts = line.pop(0)
                for item in line:
                    if self.tsList.get(item):
                        different = abs(bitVector.bit_length() - self.tsList[item].bit_length())
                        self.tsList[item] = self.tsList[item] << different
                        self.tsList[item] = self.tsList[item] | 0b1
                    else:
                        self.tsList[item] = bitVector
                count += 1
            self.tsmax = int(ts)
            for item in self.tsList:
                different = abs(bitVector.bit_length() - self.tsList[item].bit_length())
                self.tsList[item] = self.tsList[item] << different
            self.maxPer = (count - 1) * self.maxPer
            self.maxSoPer = (count - 1) * self.maxSoPer
            self.minDur = (count - 1) * self.minDur

    def generateLPP(self):
        """
        Generate local periodic items from bit vector tsList.
        When finish generating local periodic items, execute mining depth first search.
        """
        I = set()
        PTL = {}
        for item in self.tsList:
            PTL[item] = set()
            ts = list(bin(self.tsList[item]))
            ts = ts[2:]
            start = -1
            currentTs = 1
            for t in ts[currentTs:]:
                if t == '0':
                    currentTs += 1
                    continue
                else:
                    tsPre = currentTs
                    currentTs += 1
                    break
            for t in ts[currentTs:]:
                if t == '0':
                    currentTs += 1
                    continue
                else:
                    per = currentTs - tsPre
                    if per <= self.maxPer and start == -1:
                        start = tsPre
                        soPer = self.maxSoPer
                    if start != -1:
                        soPer = max(0, soPer + per - self.maxPer)
                        if soPer > self.maxSoPer:
                            if tsPre - start >= self.minDur:
                                PTL[item].add((start, tsPre))
                            """else:
                                bitVector = 0b1 << currentTs
                                different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                                bitVector = bitVector | 0b1
                                bitVector = bitVector << different
                                self.tsList[item] = self.tsList[item] | bitVector"""
                            start = -1
                    tsPre = currentTs
                    currentTs += 1
            if start != -1:
                soPer = max(0, soPer + self.tsmax - tsPre - self.maxPer)
                if soPer > self.maxSoPer and tsPre - start >= self.minDur:
                    PTL[item].add((start, tsPre))
                """else:
                    bitVector = 0b1 << currentTs+1
                    different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                    bitVector = bitVector | 0b1
                    bitVector = bitVector << different
                    self.tsList[item] = self.tsList[item] | bitVector"""
                if soPer <= self.maxSoPer and self.tsmax - start >= self.minDur:
                    PTL[item].add((start, self.tsmax))
                """else:
                    bitVector = 0b1 << currentTs+1
                    different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                    bitVector = bitVector | 0b1
                    bitVector = bitVector << different
                    self.tsList[item] = self.tsList[item] | bitVector"""
            if len(PTL[item]) > 0:
                I |= {item}
                self.finalPatterns[item] = PTL[item]
        I = sorted(list(I))
        map = {-1:I}
        I = set(I)
        while len(map) > 0:
            map = self.LPPMBreadthSearch(map)

    def calclatePTL(self, tsList):
        """
        calculate PTL from tslist as bit vector.
        :param tsList: it is one item's tslist which is used bit vector.
        :type tsList: int
        :return: it is PTL of input item.
        """
        tsList = list(bin(tsList))
        tsList = tsList[2:]
        start = -1
        currentTs = 1
        PTL = set()
        for ts in tsList[currentTs:]:
            if ts == '0':
                currentTs += 1
                continue
            else:
                tsPre = currentTs
                currentTs += 1
                break
        for ts in tsList[currentTs:]:
            if ts == '0':
                currentTs += 1
                continue
            else:
                per = currentTs - tsPre
                if per <= self.maxPer and start == -1:
                    start = tsPre
                    soPer = self.maxSoPer
                if start != -1:
                    soPer = max(0, soPer + per - self.maxPer)
                    if soPer > self.maxSoPer:
                        if tsPre - start >= self.minDur:
                            PTL.add((start, tsPre))
                        start = -1
                tsPre = currentTs
                currentTs += 1
        if start != -1:
            soPer = max(0, soPer + self.tsmax - tsPre - self.maxPer)
            if soPer > self.maxSoPer and tsPre - start >= self.minDur:
                PTL.add((start, tsPre))
            if soPer <= self.maxSoPer and self.tsmax - start >= self.minDur:
                PTL.add((start, tsPre))
        return PTL

    def LPPMBreadthSearch(self, wmap):
        """
        Mining n-length local periodic pattens from n-1-length patterns by depth first search.
        :param wmap: it is w length patterns and its conditional items
        :type wmap: dict
        :return w1map: it is w+1 length patterns and its conditional items
        :rtype w1map: dict
        """
        w1map = {}
        for p in wmap:
            if p != -1:
                listP = p
                if type(p) == str:
                    listP = [p]
                tsp = self.tsList[listP[0]]
                for item in listP[1:]:
                    tsp = tsp & self.tsList[item]

            for x in range(len(wmap[p])-1):
                for y in range(x+1,len(wmap[p])):
                    if p == -1:
                        tspxy = self.tsList[wmap[p][x]] & self.tsList[wmap[p][y]]
                    else:
                        tspxy = tsp & self.tsList[wmap[p][x]] & self.tsList[wmap[p][y]]
                    PTL = self.calclatePTL(tspxy)
                    if len(PTL) > 0:
                        if p == -1:
                            if not w1map.get(wmap[p][x]):
                                w1map[wmap[p][x]] = []
                            pattern = (wmap[p][x], wmap[p][y])
                            self.finalPatterns[pattern] = PTL
                            w1map[wmap[p][x]].append(wmap[p][y])
                        else:
                            pattern = [item for item in listP]
                            pattern.append(wmap[p][x])
                            pattern1 = pattern.copy()
                            pattern.append(wmap[p][y])
                            self.finalPatterns[tuple(pattern)] = PTL
                            if not w1map.get(tuple(pattern1)):
                                w1map[tuple(pattern1)] = []
                            w1map[tuple(pattern1)].append(wmap[p][y])
        return w1map


    def startMine(self):
        """
        Mining process start from here.
        """
        self.startTime = time.time()
        self.createTSlist()
        self.generateLPP()
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsInDataFrame(self):
        """Storing final local periodic patterns in a dataframe

        :return: returning local periodic patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'PTL'])
        return dataFrame

    def storePatternsInFile(self, outFile):
        """Complete set of local periodic patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            writer.write(f'{x} : {y}\n')
            # patternsAndPTL = x + ":" + str(y)
            # writer.write("%s \n" % patternsAndPTL)

    def getLocalPeriodicPatterns(self):
        """ Function to send the set of local periodic patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns

if __name__ == '__main__':
    if len(sys.argv) == 6:
        ap = LPPMBreadth(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.startMine()
        localPeriodicPatterns = ap.getLocalPeriodicPatterns()
        print(f"Total number of Frequent Patterns: {len(localPeriodicPatterns)}")
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print(f'Total Memory in USS: {memUSS}')
        memRSS = ap.getMemoryRSS()
        print(f'Total Memory in RSS: {memRSS}')
        run = ap.getRuntime()
        print(f'Total ExecutionTime in seconds: {run}')
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



from PAMI.frequentSpatialPattern.basic.abstract import *
import sys
import pandas as pd


class SpatialEclat(spatialFrequentPatterns):
    """ 
        Spatial Eclat is a Extension of ECLAT algorithm,which  stands for Equivalence Class Clustering and bottom-up
        Lattice Traversal.It is one of the popular methods of Association Rule mining. It is a more efficient and
        scalable version of the Apriori algorithm.

            ...

    Attributes :
    ----------
            iFile : str
                Input file name or path of the input file
            nFile: str:
               Name of Neighbourhood file name
            minSup: float
                UserSpecified minimum support value. It has to be given in terms of count of total number of
                transactions in the input database/file
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            oFile : str
                Name of the output file to store complete set of frequent patterns
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the complete set of transactions available in the input database/file

    Methods :
    -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            storePatternsInFile(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsInDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            findDelimiter(line)
                Identifying the delimiter of the input file
            creatingItemSets(iFileName)
                Storing the complete transactions of the database/input file in a database variable
            frequentOneItem()
                Generating one frequent patterns
            dictKeysToInt(iList)
                Converting dictionary keys to integer elements
            eclatGeneration(cList)
                It will generate the combinations of frequent items
            generatespatialFrequentPatterns(tidList)
                It will generate the combinations of frequent items from a list of items
            convert(value):
                To convert the given user specified value    
            getNighboirItems(keyset):
                A function to get commom neighbours of a itemset
             mapNighbours(file):
                A function to map items to their neighbours

    Executing the code on terminal :
    ------------------------------
            Format: python3 SpatialEclat.py <inputFile> <outputFile> <neighbourFile> <minSup>
            Examples:
            python3 SpatialEclat.py sampleTDB.txt output.txt sampleN.txt 0.5 (minSup will be considered in percentage of database transactions)
            python3 SpatialEclat.py sampleTDB.txt output.txt sampleN.txt 3 (minSup will be considered in support count or frequency)
                                                                (it considers "\t" as separator)
                    SpatialEclat.py sampleTDB.txt output.txt sampleN.txt 3 , 
                                                                (it will consider "," as a separator)

    Sample run of importing the code :
    -------------------------------
        import SpatialEclat as alg
        obj = alg.SpatialEclat("sampleTDB.txt", "sampleN.txt", 5)

        obj.startMine()

        spatialFrequentPatterns = obj.getPatterns()

        print("Total number of Spatial Frequent Patterns:", len(spatialFrequentPatterns))

        obj.storePatternsInFile("outFile")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)


    Credits:
    -------
        The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
    """

    minSup = float()
    startTime = float()
    endTime = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    nFile=" "
    memoryUSS = float()
    memoryRSS = float()
    Database = []
    sep="\t"

    def __init__(self, iFile, nFile, minSup, sep="\t"):
        super().__init__(iFile, nFile, minSup, sep)
        self.NighboursMap = {}

    def creatingItemSets(self, iFileName):
        """Storing the complete transactions of the database/input file in a database variable

            :param iFileName: user given input file/input file path
            :type iFileName: str
            """
        self.Database = []
        lineNumber = 0
        with open(iFileName, 'r', encoding='utf-8') as f:
            for line in f:
                # line.strip()
                if lineNumber == 0:
                    lineNumber += 1
                    # delimiter = self.findDelimiter([*line])
                    # li=[lineNumber]
                    li = line.split(self.sep)
                    li1 = [i.rstrip() for i in li]
                    self.Database.append([i.rstrip() for i in li1])
                    # else:
                    # self.Database.append(li)
                    # data.append([lineNumber,li1])
                else:
                    lineNumber += 1
                    li = line.split(self.sep)
                    # if delimiter==',':
                    li1 = [i.rstrip() for i in li]
                    self.Database.append(li1)

    # function to get frequent one pattern
    def frequentOneItem(self):
        """Generating one frequent patterns"""

        candidate = {}
        # global finalPatterns, minSup, Database
        # self.minSup = self.minSup
        for i in range(len(self.Database)):
            for j in range(len(self.Database[i])):
                if self.Database[i][j] not in candidate:
                    candidate[self.Database[i][j]] = [i]
                else:
                    candidate[self.Database[i][j]] += [i]
        self.finalPatterns = {keys: value for keys, value in candidate.items() if len(value) >= self.minSup}
        # print(candidate)

    def convert(self, value):
        """
        To convert the given user specified value
        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.Database) * value)
            else:
                value = int(value)
        return value

    @staticmethod
    def dictKeysToInt(iList):
        """Converting dictionary keys to integer elements

        :param iList: Dictionary with patterns as keys and their support count as a value
        :type iList: dict
        :returns: list of integer patterns to represent dictionary keys
        :rtype: list
        """

        temp = []
        for ite in iList.keys():
            ite = [int(i) for i in ite.strip('[]').split(',')]
            temp.append(ite)
            # print(sorted(temp))
        return sorted(temp)

    def eclatGeneration(self, cList):
        """It will generate the combinations of frequent items

        :param cList :it represents the items with their respective transaction identifiers
        :type cList: dictionary
        :return: returning transaction dictionary
        :rtype: dict
        """
        # to generate all
        tidList = {}
        key = list(cList.keys())
        for i in range(0, len(key)):
            nighbousItems = self.getNighboirItems(key[i])
            for j in range(i + 1, len(key)):
                # print(c[key[i]],c[key[j]])
                if not key[j] in nighbousItems:
                    continue
                intersectionList = list(set(cList[key[i]]).intersection(set(cList[key[j]])))
                itemList = []
                itemList += key[i]
                itemList += key[j]
                if len(intersectionList) >= self.minSup:
                    itemList.sort()
                    if tuple(itemList) not in tidList:
                        tidList[tuple(set(itemList))] = intersectionList
        return tidList

    def generatespatialFrequentPatterns(self, tidList):
        """It will generate the combinations of frequent items from a list of items

        :param tidList :it represents the items with their respective transaction identifiers
        :type tidList: dictionary
        :return: returning transaction dictionary
        :rtype: dict
        """
        tidList1 = {}
        if len(tidList) == 0:
            print("There are no more candidate sets")
        else:
            key = list(tidList.keys())
            for i in range(0, len(key)):
                nighbousItems = self.getNighboirItems(key[i])
                for j in range(i + 1, len(key)):
                    if not key[j] in nighbousItems:
                        continue
                    intersectionList = list(set(tidList[key[i]]).intersection(set(tidList[key[j]])))
                    itemList = []
                    if len(intersectionList) >= self.minSup:
                        itemList += key[i], key[j]
                        itemList.sort()
                        tidList1[tuple(itemList)] = intersectionList

        return tidList1

    def getNighboirItems(self, keyset):
        """
            A function to get Neighbours of a item
            :param keyset:itemset
            :type keyset:str or tuple
            :return: set of common neighbours 
            :rtype:set
        """
        itemNibours = self.NighboursMap.keys()
        if isinstance(keyset, str):
            if self.NighboursMap.get(keyset) is None:
                return []
            itemNibours = list(set(itemNibours).intersection(set(self.NighboursMap.get(keyset))))
        if isinstance(keyset, tuple):
            keyset = list(keyset)
            # print(keyset)
            for j in range(0, len(keyset)):
                i = keyset[j]
                itemNibours = list(set(itemNibours).intersection(set(self.NighboursMap.get(i))))
        return itemNibours

    def mapNighbours(self, name):
        """
            A function to map items to their Neighbours
            :param name: item name
            :type name: int
        """
        with open(name, 'r', encoding='utf-8') as f:
            for line in f:
                li = line.split(self.sep)
                item = li[0]
                nibs = li[1:]
                self.NighboursMap[item] = nibs

    def startMine(self):
        """Frequent pattern mining process will start from here"""

        # global items_sets, endTime, startTime
        self.startTime = time.time()
        if self.iFile is None:
            raise Exception("Please enter the file path or file name:")
        iFileName = self.iFile
        self.creatingItemSets(iFileName)
        self.minSup = self.convert(self.minSup)
        self.mapNighbours(self.nFile)
        print(self.minSup)
        self.frequentOneItem()
        frequentSet = self.generatespatialFrequentPatterns(self.finalPatterns)
        # print("frequentSet",self.finalPatterns)
        for x, y in frequentSet.items():
            if x not in self.finalPatterns:
                self.finalPatterns[x] = y
        while 1:
            frequentSet = self.eclatGeneration(frequentSet)
            # print("frequentSet",frequentSet)
            for x, y in frequentSet.items():
                if x not in self.finalPatterns:
                    self.finalPatterns[x] = y
            if len(frequentSet) == 0:
                break
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("Spatial Frequent patterns were generated successfully using SpatialEclat algorithm")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsInDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def storePatternsInFile(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            pat = ""
            for i in x:
                pat += str(i) + " "
            patternsAndSupport = pat + ": " + str(len(y))
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6:
            ap = SpatialEclat(sys.argv[1], sys.argv[3], sys.argv[4],sys.argv[5])
        if len(sys.argv) == 5:
            ap = SpatialEclat(sys.argv[1], sys.argv[3], sys.argv[4])
        ap.startMine()
        spatialFrequentPatterns = ap.getPatterns()
        print("Total number of Spatial Frequent Patterns:", len(spatialFrequentPatterns))
        ap.storePatternsInFile(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in seconds:", run)
    else:
        print("Error!,Parameter doesn't match")